==> org-lsst-ccs-subsystem-utility-main-1.4.4-sources.jar: org/lsst/ccs/subsystem/utility/VpcControlUT.java <==
package org.lsst.ccs.subsystem.utility;

import java.time.Duration;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.logging.Logger;
import org.lsst.ccs.Subsystem;
import org.lsst.ccs.commons.annotations.ConfigurationParameter;
import org.lsst.ccs.commons.annotations.LookupField;
import org.lsst.ccs.commons.annotations.LookupName;
import org.lsst.ccs.commons.annotations.LookupPath;
import org.lsst.ccs.framework.AgentPeriodicTask;
import org.lsst.ccs.framework.HasLifecycle;
import org.lsst.ccs.monitor.Channel;
import org.lsst.ccs.services.AgentPeriodicTaskService;
import org.lsst.ccs.subsystem.common.ErrorUtils;
import org.lsst.ccs.subsystem.utility.constants.UtilTrunkHeaters;
import org.lsst.ccs.subsystem.utility.data.UtilityException;

/**
 *  Implements a temperature controller for the volume purge cabinet.
 *
 *  @author Owen Saxton
 */
public class VpcControlUT implements HasLifecycle {

    private static final String VPC = "Vpc";
    private static final Map<Integer, Integer> heaterChans = new HashMap<>();
    static {
        heaterChans.put(UtilTrunkHeaters.HEATER_VPC1_ID, BfrDeviceUT.CHAN_HEATER_1);
        heaterChans.put(UtilTrunkHeaters.HEATER_VPC2_ID, BfrDeviceUT.CHAN_HEATER_2);
    }

    @LookupField(strategy = LookupField.Strategy.TOP)
    Subsystem subsys;
    @LookupName
    private String name;
    @LookupPath
    private String path;
    @LookupField(strategy = LookupField.Strategy.TREE)
    private AgentPeriodicTaskService periodicTaskServices;

    @LookupField(strategy = LookupField.Strategy.TREE)
    private BfrDeviceUT bfrDevc;
    
    @ConfigurationParameter(category=VPC, maxLength=UtilTrunkHeaters.NUM_VPC_HEATERS, isFinal=true)
    private volatile double[] deltaTemps = new double[0];
    @ConfigurationParameter(category=VPC, isFinal=true)
    private volatile Integer updateTime = Integer.MAX_VALUE; // The update time interval (msec)

    private List<Channel> refTempChans;   // Reference temperature channels to use
    private List<Channel> ctrlTempChans;  // Control temperature channels to use

    private static final Logger LOG = Logger.getLogger(VpcControlUT.class.getName());
    private boolean active = false;    // True if loop is active
    private boolean loopFailedRef = false, loopFailedCtrl = false, loopFailedHtr = false;
    private long lastLoopTime;


    /**
     *  Sets up the VPC temperature control loop timer task.
     */
    @Override
    public void build() {
        AgentPeriodicTask apt = new AgentPeriodicTask("vpc-loop-" + name, () -> iterateLoop()).withPeriod(Duration.ofMillis(500));
        periodicTaskServices.scheduleAgentPeriodicTask(apt);
    }


    /**
     *  Initializes the parameters.
     */
    @Override
    public void init()
    {
        if (bfrDevc == null) {
            ErrorUtils.reportConfigError(LOG, path, "BfrDevice component", "is missing");
        }
        if (refTempChans == null || refTempChans.isEmpty()) {
            ErrorUtils.reportConfigError(LOG, path, "refTempChans", "is missing or empty");
        }
        if (ctrlTempChans == null || ctrlTempChans.isEmpty()) {
            ErrorUtils.reportConfigError(LOG, path, "ctrlTempChans", "is missing or empty");
        }
        if (deltaTemps.length != UtilTrunkHeaters.NUM_VPC_HEATERS) {
            ErrorUtils.reportConfigError(LOG, path, "deltaTemps", "must be of length " + UtilTrunkHeaters.NUM_VPC_HEATERS);
        }
        if (updateTime == Integer.MAX_VALUE) {
            ErrorUtils.reportConfigError(LOG, path, "updateTime", "is missing");
        }
    }


    /**
     *  Sets a target delta temperature.
     *
     *  @param  heater  The VPC heater ID
     *  @param  temp    The temperature to set
     *  @throws  UtilityException
     */
    public void setDeltaTemp(int heater, double temp) throws UtilityException
    {
        deltaTemps[heater] = temp;
    }


    /**
     *  Gets a target delta temperature.
     *
     *  @param  heater  The VPC heater ID
     *  @return  The set temperature
     */
    public double getDeltaTemp(int heater)
    {
        return deltaTemps[heater];
   }


    /**
     *  Starts the control loop.
     *
     *  @throws  UtilityException
     */
    public void startLoop() throws UtilityException
    {
        active = true;
    }


    /**
     *  Stops the control loop.
     *
     *  @throws  UtilityException
     */
    public void stopLoop() throws UtilityException
    {
        active = false;
    }


    /**
     *  Gets the control loop state.
     *
     *  @return  Whether the control loop is active
     */
    public boolean isLoopActive()
    {
        return active;
    }


    /**
     *  Timer routine for control loop iteration.
     */
    private void iterateLoop()
    {
        if (!active) return;
        long time = System.currentTimeMillis();
        if (time < lastLoopTime + updateTime) return;
        lastLoopTime = time;
        double refTemp = 0.0, ctrlTemp = 0.0;
        int count = 0;
        for (Channel tempChan : refTempChans) {
            double value = tempChan.getValue();
            if (!Double.isNaN(value)) {
                refTemp += value;
                count++;
            }
        }
        if (count > 0) {
            refTemp /= count;
            loopFailedRef = false;
        }
        else {
            if (!loopFailedRef) {
                LOG.severe("Control loop iteration failed: no valid reference temperature values available");
                loopFailedRef = true;
            }
            return;
        }
        count = 0;
        for (Channel tempChan : ctrlTempChans) {
            double value = tempChan.getValue();
            if (!Double.isNaN(value)) {
                ctrlTemp += value;
                count++;
            }
        }
        if (count > 0) {
            ctrlTemp /= count;
            loopFailedCtrl = false;
        }
        else {
            if (!loopFailedCtrl) {
                LOG.severe("Control loop iteration failed: no valid control temperature values available");
                loopFailedCtrl = true;
            }
            return;
        }
        if (!setHeaters(refTemp - ctrlTemp)) {
            if (!loopFailedHtr) {
                LOG.severe("Control loop iteration failed: heaters are offline");
                loopFailedHtr = true;
            }
            return;
        }
        loopFailedHtr = false;
    }


    /**
     *  Sets the heater state
     *
     *  @param  deltaTemp  The measured temperature difference
     *  @return  Whether heaters are online
     */
    private boolean setHeaters(double deltaTemp)
    {
        for (int htrId = 0; htrId < UtilTrunkHeaters.NUM_VPC_HEATERS; htrId++) {
            int htrChan = heaterChans.get(htrId);
            Boolean htrOn = bfrDevc.isSwitchOn(htrChan);
            if (htrOn == null) return false;
            if (deltaTemp > deltaTemps[htrId]) {
                if (!htrOn) {
                    bfrDevc.switchOn(htrChan);
                }
            }
            else if (deltaTemp <= 0.0) {
                if (htrOn) {
                    bfrDevc.switchOff(htrChan);
                }
            }
        }
        return true;
    }

}

==> org-lsst-ccs-subsystem-utility-main-1.4.11-sources.jar: org/lsst/ccs/subsystem/utility/VpcControlUT.java <==
package org.lsst.ccs.subsystem.utility;

import java.time.Duration;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.logging.Logger;
import org.lsst.ccs.Subsystem;
import org.lsst.ccs.commons.annotations.ConfigurationParameter;
import org.lsst.ccs.commons.annotations.LookupField;
import org.lsst.ccs.commons.annotations.LookupName;
import org.lsst.ccs.commons.annotations.LookupPath;
import org.lsst.ccs.framework.AgentPeriodicTask;
import org.lsst.ccs.framework.HasLifecycle;
import org.lsst.ccs.monitor.Channel;
import org.lsst.ccs.services.AgentPeriodicTaskService;
import org.lsst.ccs.subsystem.common.ErrorUtils;
import org.lsst.ccs.subsystem.utility.constants.UtilTrunkHeaters;
import org.lsst.ccs.subsystem.utility.data.UtilityException;

/**
 *  Implements a temperature controller for the volume purge cabinet.
 *
 *  @author Owen Saxton
 */
public class VpcControlUT implements HasLifecycle {

    private static final String VPC = "Vpc";
    private static final Map<Integer, Integer> heaterChans = new HashMap<>();
    static {
        heaterChans.put(UtilTrunkHeaters.HEATER_VPC1_ID, BfrDeviceUT.CHAN_HEATER_1);
        heaterChans.put(UtilTrunkHeaters.HEATER_VPC2_ID, BfrDeviceUT.CHAN_HEATER_2);
    }

    @LookupField(strategy = LookupField.Strategy.TOP)
    Subsystem subsys;
    @LookupName
    private String name;
    @LookupPath
    private String path;
    @LookupField(strategy = LookupField.Strategy.TREE)
    private AgentPeriodicTaskService periodicTaskServices;

    @LookupField(strategy = LookupField.Strategy.TREE)
    private BfrDeviceUT bfrDevc;
    
    @ConfigurationParameter(category=VPC, maxLength=UtilTrunkHeaters.NUM_VPC_HEATERS, isFinal=true, units = "Celsius", description="Heater turn-on temperature differences")
    private volatile double[] deltaTemps = new double[0];
    @ConfigurationParameter(category=VPC, isFinal=true, units = "ms", description="Control loop update time")
    private volatile Integer updateTime = Integer.MAX_VALUE; // The update time interval (msec)

    private List<Channel> refTempChans;   // Reference temperature channels to use
    private List<Channel> ctrlTempChans;  // Control temperature channels to use

    private static final Logger LOG = Logger.getLogger(VpcControlUT.class.getName());
    private boolean active = false;    // True if loop is active
    private boolean loopFailedRef = false, loopFailedCtrl = false, loopFailedHtr = false;
    private long lastLoopTime;


    /**
     *  Sets up the VPC temperature control loop timer task.
     */
    @Override
    public void build() {
        AgentPeriodicTask apt = new AgentPeriodicTask("vpc-loop-" + name, () -> iterateLoop()).withPeriod(Duration.ofMillis(500));
        periodicTaskServices.scheduleAgentPeriodicTask(apt);
    }


    /**
     *  Initializes the parameters.
     */
    @Override
    public void init()
    {
        if (bfrDevc == null) {
            ErrorUtils.reportConfigError(LOG, path, "BfrDevice component", "is missing");
        }
        if (refTempChans == null || refTempChans.isEmpty()) {
            ErrorUtils.reportConfigError(LOG, path, "refTempChans", "is missing or empty");
        }
        if (ctrlTempChans == null || ctrlTempChans.isEmpty()) {
            ErrorUtils.reportConfigError(LOG, path, "ctrlTempChans", "is missing or empty");
        }
        if (deltaTemps.length != UtilTrunkHeaters.NUM_VPC_HEATERS) {
            ErrorUtils.reportConfigError(LOG, path, "deltaTemps", "must be of length " + UtilTrunkHeaters.NUM_VPC_HEATERS);
        }
        if (updateTime == Integer.MAX_VALUE) {
            ErrorUtils.reportConfigError(LOG, path, "updateTime", "is missing");
        }
    }


    /**
     *  Sets a target delta temperature.
     *
     *  @param  heater  The VPC heater ID
     *  @param  temp    The temperature to set
     *  @throws  UtilityException
     */
    public void setDeltaTemp(int heater, double temp) throws UtilityException
    {
        deltaTemps[heater] = temp;
    }


    /**
     *  Gets a target delta temperature.
     *
     *  @param  heater  The VPC heater ID
     *  @return  The set temperature
     */
    public double getDeltaTemp(int heater)
    {
        return deltaTemps[heater];
   }


    /**
     *  Starts the control loop.
     *
     *  @throws  UtilityException
     */
    public void startLoop() throws UtilityException
    {
        active = true;
    }


    /**
     *  Stops the control loop.
     *
     *  @throws  UtilityException
     */
    public void stopLoop() throws UtilityException
    {
        active = false;
    }


    /**
     *  Gets the control loop state.
     *
     *  @return  Whether the control loop is active
     */
    public boolean isLoopActive()
    {
        return active;
    }


    /**
     *  Timer routine for control loop iteration.
     */
    private void iterateLoop()
    {
        if (!active) return;
        long time = System.currentTimeMillis();
        if (time < lastLoopTime + updateTime) return;
        lastLoopTime = time;
        double refTemp = 0.0, ctrlTemp = 0.0;
        int count = 0;
        for (Channel tempChan : refTempChans) {
            double value = tempChan.getValue();
            if (!Double.isNaN(value)) {
                refTemp += value;
                count++;
            }
        }
        if (count > 0) {
            refTemp /= count;
            loopFailedRef = false;
        }
        else {
            if (!loopFailedRef) {
                LOG.severe("Control loop iteration failed: no valid reference temperature values available");
                loopFailedRef = true;
            }
            return;
        }
        count = 0;
        for (Channel tempChan : ctrlTempChans) {
            double value = tempChan.getValue();
            if (!Double.isNaN(value)) {
                ctrlTemp += value;
                count++;
            }
        }
        if (count > 0) {
            ctrlTemp /= count;
            loopFailedCtrl = false;
        }
        else {
            if (!loopFailedCtrl) {
                LOG.severe("Control loop iteration failed: no valid control temperature values available");
                loopFailedCtrl = true;
            }
            return;
        }
        if (!setHeaters(refTemp - ctrlTemp)) {
            if (!loopFailedHtr) {
                LOG.severe("Control loop iteration failed: heaters are offline");
                loopFailedHtr = true;
            }
            return;
        }
        loopFailedHtr = false;
    }


    /**
     *  Sets the heater state
     *
     *  @param  deltaTemp  The measured temperature difference
     *  @return  Whether heaters are online
     */
    private boolean setHeaters(double deltaTemp)
    {
        for (int htrId = 0; htrId < UtilTrunkHeaters.NUM_VPC_HEATERS; htrId++) {
            int htrChan = heaterChans.get(htrId);
            Boolean htrOn = bfrDevc.isSwitchOn(htrChan);
            if (htrOn == null) return false;
            if (deltaTemp > deltaTemps[htrId]) {
                if (!htrOn) {
                    bfrDevc.switchOn(htrChan);
                }
            }
            else if (deltaTemp <= 0.0) {
                if (htrOn) {
                    bfrDevc.switchOff(htrChan);
                }
            }
        }
        return true;
    }

}
